#
# This code is auto-generated from CLIc 'cle::tier1.hpp' file, do not edit manually.
#

import importlib
import warnings
from typing import Optional

import numpy as np

from ._array import Image
from ._core import Device
from ._decorators import plugin_function

clic = importlib.import_module('._pyclesperanto', package='pyclesperanto')

@plugin_function(categories=["filter", "in assistant"])
def absolute(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Computes the absolute value of every individual pixel x in a given image.
    <pre>f(x) = |x| </pre>

    Parameters
    ----------
    input_image: Image 
        The input image to be processed.
    output_image: Optional[Image] (= None)
        The output image where results are written into.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_absolute
    """
    return clic._absolute(device, input_image, output_image)

@plugin_function(categories=["combine", "in assistant"])
def add_images_weighted(
    input_image0: Image,
    input_image1: Image,
    output_image: Optional[Image] =None,
    factor0: float =1,
    factor1: float =1,
    device: Optional[Device] =None
) -> Image:
    """Calculates the sum of pairs of pixels x and y from images X and Y weighted with
    factors a and b. <pre>f(x, y, a, b) = x * a + y * b</pre>

    Parameters
    ----------
    input_image0: Image 
        The first input image to added.
    input_image1: Image 
        The second image to be added.
    output_image: Optional[Image] (= None)
        The output image where results are written into.
    factor0: float (= 1)
        Multiplication factor of each pixel of src0 before adding it.
    factor1: float (= 1)
        Multiplication factor of each pixel of src1 before adding it.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_addImagesWeighted
    """
    return clic._add_images_weighted(device, input_image0, input_image1, output_image, float(factor0), float(factor1))

@plugin_function(categories=["filter", "in assistant"])
def add_image_and_scalar(
    input_image: Image,
    output_image: Optional[Image] =None,
    scalar: float =1,
    device: Optional[Device] =None
) -> Image:
    """Adds a scalar value s to all pixels x of a given image X. <pre>f(x, s) = x +
    s</pre>

    Parameters
    ----------
    input_image: Image 
        The input image where scalare should be added.
    output_image: Optional[Image] (= None)
        The output image where results are written into.
    scalar: float (= 1)
        The constant number which will be added to all pixels.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_addImageAndScalar
    """
    return clic._add_image_and_scalar(device, input_image, output_image, float(scalar))

@plugin_function(categories=["combine", "binary processing", "in assistant", "combine labels", "label processing"])
def binary_and(
    input_image0: Image,
    input_image1: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Computes a binary image (containing pixel values 0 and 1) from two images X and
    Y by connecting pairs of pixels x and y with the binary AND operator &. All
    pixel values except 0 in the input images are interpreted as 1. <pre>f(x, y) = x
    & y</pre>

    Parameters
    ----------
    input_image0: Image 
        The first binary input image to be processed.
    input_image1: Image 
        The second binary input image to be processed.
    output_image: Optional[Image] (= None)
        The output image where results are written into.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_binaryAnd
    """
    return clic._binary_and(device, input_image0, input_image1, output_image)

@plugin_function(categories=["binary processing", "label processing", "in assistant", "bia-bob-suggestion"])
def binary_edge_detection(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Determines pixels/voxels which are on the surface of binary objects and sets
    only them to 1 in the destination image. All other pixels are set to 0.

    Parameters
    ----------
    input_image: Image 
        The binary input image where edges will be searched.
    output_image: Optional[Image] (= None)
        The output image where edge pixels will be 1.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_binaryEdgeDetection
    """
    return clic._binary_edge_detection(device, input_image, output_image)

@plugin_function(categories=["binary processing", "filter", "label processing", "in assistant", "bia-bob-suggestion"])
def binary_not(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Computes a binary image (containing pixel values 0 and 1) from an image X by
    negating its pixel values x using the binary NOT operator ! All pixel values
    except 0 in the input image are interpreted as 1. <pre>f(x) = !x</pre>

    Parameters
    ----------
    input_image: Image 
        The binary input image to be inverted.
    output_image: Optional[Image] (= None)
        The output image where results are written into.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_binaryNot
    """
    return clic._binary_not(device, input_image, output_image)

@plugin_function(categories=["combine", "binary processing", "in assistant", "combine labels", "label processing"])
def binary_or(
    input_image0: Image,
    input_image1: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Computes a binary image (containing pixel values 0 and 1) from two images X and
    Y by connecting pairs of pixels x and y with the binary OR operator |. All pixel
    values except 0 in the input images are interpreted as 1.<pre>f(x, y) = x |
    y</pre>

    Parameters
    ----------
    input_image0: Image 
        The first binary input image to be processed.
    input_image1: Image 
        The second binary input image to be processed.
    output_image: Optional[Image] (= None)
        The output image where results are written into.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_binaryOr
    """
    return clic._binary_or(device, input_image0, input_image1, output_image)

@plugin_function(categories=["combine", "binary processing", "in assistant", "combine labels", "label processing"])
def binary_subtract(
    input_image0: Image,
    input_image1: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Subtracts one binary image from another.

    Parameters
    ----------
    input_image0: Image 
        The first binary input image to be processed.
    input_image1: Image 
        The second binary input image to be subtracted from the first.
    output_image: Optional[Image] (= None)
        The output image where results are written into.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_binarySubtract
    """
    return clic._binary_subtract(device, input_image0, input_image1, output_image)

@plugin_function(categories=["combine", "binary processing", "in assistant", "combine labels", "label processing"])
def binary_xor(
    input_image0: Image,
    input_image1: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Computes a binary image (containing pixel values 0 and 1) from two images X and
    Y by connecting pairs of pixels x and y with the binary operators AND &, OR |
    and NOT ! implementing the XOR operator. All pixel values except 0 in the input
    images are interpreted as 1. <pre>f(x, y) = (x & !y) | (!x & y)</pre>

    Parameters
    ----------
    input_image0: Image 
        The first binary input image to be processed.
    input_image1: Image 
        The second binary input image to be processed.
    output_image: Optional[Image] (= None)
        The output image where results are written into.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_binaryXOr
    """
    return clic._binary_xor(device, input_image0, input_image1, output_image)

@plugin_function(categories=["filter", "binary processing"])
def binary_supinf(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Compute the maximum of the erosion with plannar structuring elements. Warning:
    This operation is only supported BINARY data type images.

    Parameters
    ----------
    input_image: Image 
        The binary input image to be processed.
    output_image: Optional[Image] (= None)
        The output image where results are written into.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._binary_supinf(device, input_image, output_image)

@plugin_function(categories=["filter", "binary processing"])
def binary_infsup(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Compute the minimum of the dilation with plannar structuring elements. Warning:
    This operation is only supported BINARY data type images.

    Parameters
    ----------
    input_image: Image 
        The binary input image to be processed.
    output_image: Optional[Image] (= None)
        The output image where results are written into.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._binary_infsup(device, input_image, output_image)

@plugin_function
def block_enumerate(
    input_image0: Image,
    input_image1: Image,
    output_image: Optional[Image] =None,
    blocksize: int =256,
    device: Optional[Device] =None
) -> Image:
    """Enumerates pixels with value 1 in a onedimensional image For example handing
    over the image [0, 1, 1, 0, 1, 0, 1, 1] would be processed to an image [0, 1, 2,
    0, 3, 0, 4, 5] This functionality is important in connected component neccessary
    (see also sum_reduction_x). In the above example, with blocksize 4, that would
    be the sum array: [2, 3] labeling. Processing is accelerated by paralellization
    in blocks. Therefore, handing over precomputed block sums is Note that the block
    size when calling this function and sum_reduction must be identical

    Parameters
    ----------
    input_image0: Image 
        input binary vector image
    input_image1: Image 
        precomputed sums of blocks
    output_image: Optional[Image] (= None)
        output enumerated vector image
    blocksize: int (= 256)
        
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._block_enumerate(device, input_image0, input_image1, output_image, int(blocksize))

@plugin_function(categories=["filter", "combine", "in assistant"])
def convolve(
    input_image0: Image,
    input_image1: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Convolve the image with a given kernel image. It is recommended that the kernel
    image has an odd size in X, Y and Z.

    Parameters
    ----------
    input_image0: Image 
        First input image to process.
    input_image1: Image 
        Second input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_convolve
    """
    return clic._convolve(device, input_image0, input_image1, output_image)

@plugin_function
def copy(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Copies an image. <pre>f(x) = x</pre>

    Parameters
    ----------
    input_image: Image 
        Input image to copy.
    output_image: Optional[Image] (= None)
        Output copy image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_copy
    """
    return clic._copy(device, input_image, output_image)

@plugin_function
def copy_slice(
    input_image: Image,
    output_image: Optional[Image] =None,
    slice: int =0,
    device: Optional[Device] =None
) -> Image:
    """This method has two purposes: It copies a 2D image to a given slice z position
    in a 3D image stack or It copies a given slice at position z in an image stack
    to a 2D image. The first case is only available via ImageJ macro. If you are
    using it, it is recommended that the target 3D image already preexists in GPU
    memory before calling this method. Otherwise, CLIJ create the image stack with z
    planes.

    Parameters
    ----------
    input_image: Image 
        Input image to copy from.
    output_image: Optional[Image] (= None)
        Output copy image slice.
    slice: int (= 0)
        
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_copySlice
    """
    return clic._copy_slice(device, input_image, output_image, int(slice))

@plugin_function
def copy_horizontal_slice(
    input_image: Image,
    output_image: Optional[Image] =None,
    slice: int =0,
    device: Optional[Device] =None
) -> Image:
    """This method has two purposes: It copies a 2D image to a given slice y position
    in a 3D image stack or It copies a given slice at position y in an image stack
    to a 2D image.

    Parameters
    ----------
    input_image: Image 
        Input image to copy from.
    output_image: Optional[Image] (= None)
        Output copy image slice.
    slice: int (= 0)
        
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_copySlice
    """
    return clic._copy_horizontal_slice(device, input_image, output_image, int(slice))

@plugin_function
def copy_vertical_slice(
    input_image: Image,
    output_image: Optional[Image] =None,
    slice: int =0,
    device: Optional[Device] =None
) -> Image:
    """This method has two purposes: It copies a 2D image to a given slice x position
    in a 3D image stack or It copies a given slice at position x in an image stack
    to a 2D image.

    Parameters
    ----------
    input_image: Image 
        Input image to copy from.
    output_image: Optional[Image] (= None)
        Output copy image slice.
    slice: int (= 0)
        
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_copySlice
    """
    return clic._copy_vertical_slice(device, input_image, output_image, int(slice))

@plugin_function
def crop(
    input_image: Image,
    output_image: Optional[Image] =None,
    start_x: int =0,
    start_y: int =0,
    start_z: int =0,
    width: int =1,
    height: int =1,
    depth: int =1,
    device: Optional[Device] =None
) -> Image:
    """Crops a given substack out of a given image stack. Note: If the destination
    image preexists already, it will be overwritten and keep it's dimensions.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    start_x: int (= 0)
        Starting index coordicante x.
    start_y: int (= 0)
        Starting index coordicante y.
    start_z: int (= 0)
        Starting index coordicante z.
    width: int (= 1)
        Width size of the region to crop.
    height: int (= 1)
        Height size of the region to crop.
    depth: int (= 1)
        Depth size of the region to crop.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_crop3D
    """
    return clic._crop(device, input_image, output_image, int(start_x), int(start_y), int(start_z), int(width), int(height), int(depth))

@plugin_function(categories=["filter", "in assistant"])
def cubic_root(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Computes the cubic root of each pixel.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._cubic_root(device, input_image, output_image)

@plugin_function(categories=["binarize", "label processing", "in assistant", "bia-bob-suggestion"])
def detect_label_edges(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Takes a labelmap and returns an image where all pixels on label edges are set to
    1 and all other pixels to 0.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_detectLabelEdges
    """
    return clic._detect_label_edges(device, input_image, output_image)

@plugin_function(categories=["binary processing"])
def dilate_box(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Computes a binary image with pixel values 0 and 1 containing the binary dilation
    of a given input image. The dilation takes the Mooreneighborhood (8 pixels in 2D
    and 26 pixels in 3d) into account. The pixels in the input image with pixel
    value not equal to 0 will be interpreted as 1. This method is comparable to the
    'Dilate' menu in ImageJ in case it is applied to a 2D image. The only difference
    is that the output image contains values 0 and 1 instead of 0 and 255.

    Parameters
    ----------
    input_image: Image 
        Input image to process. Input image to process.
    output_image: Optional[Image] (= None)
        Output result image. Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_dilateBox
    """
    return clic._dilate_box(device, input_image, output_image)

@plugin_function(categories=["binary processing"])
def dilate_sphere(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Computes a binary image with pixel values 0 and 1 containing the binary dilation
    of a given input image. The dilation takes the vonNeumannneighborhood (4 pixels
    in 2D and 6 pixels in 3d) into account. The pixels in the input image with pixel
    value not equal to 0 will be interpreted as 1.

    Parameters
    ----------
    input_image: Image 
        Input image to process. Input image to process.
    output_image: Optional[Image] (= None)
        Output result image. Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_dilateSphere
    """
    return clic._dilate_sphere(device, input_image, output_image)

@plugin_function(categories=["binary processing"])
def dilate(
    input_image: Image,
    output_image: Optional[Image] =None,
    connectivity: str ="box",
    device: Optional[Device] =None
) -> Image:
    """Computes a binary image with pixel values 0 and 1 containing the binary dilation
    of a given input image. The dilation apply the Mooreneighborhood (8 pixels in 2D
    and 26 pixels in 3d) for the "box" connectivity and the vonNeumannneighborhood
    (4 pixels in 2D and 6 pixels in 3d) for a "sphere" connectivity. The pixels in
    the input image with pixel value not equal to 0 will be interpreted as 1.

    Parameters
    ----------
    input_image: Image 
        Input image to process. Input image to process.
    output_image: Optional[Image] (= None)
        Output result image. Output result image.
    connectivity: str (= "box")
        Element shape, "box" or "sphere".
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_dilateBox
    [2] https://clij.github.io/clij2-docs/reference_dilateSphere
    """
    return clic._dilate(device, input_image, output_image, str(connectivity))

@plugin_function(categories=["combine", "in assistant"])
def divide_images(
    input_image0: Image,
    input_image1: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Divides two images X and Y by each other pixel wise. <pre>f(x, y) = x / y</pre>

    Parameters
    ----------
    input_image0: Image 
        First input image to process.
    input_image1: Image 
        Second input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_divideImages
    """
    return clic._divide_images(device, input_image0, input_image1, output_image)

@plugin_function(categories=["filter", "in assistant"])
def divide_scalar_by_image(
    input_image: Image,
    output_image: Optional[Image] =None,
    scalar: float =0,
    device: Optional[Device] =None
) -> Image:
    """Divides a scalar by an image pixel by pixel. <pre>f(x, s) = s / x</pre>

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    scalar: float (= 0)
        
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._divide_scalar_by_image(device, input_image, output_image, float(scalar))

@plugin_function(categories=["combine", "binarize", "in assistant"])
def equal(
    input_image0: Image,
    input_image1: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Determines if two images A and B equal pixel wise. <pre>f(a, b) = 1 if a == b; 0
    otherwise.</pre>

    Parameters
    ----------
    input_image0: Image 
        The first image to be compared with.
    input_image1: Image 
        The second image to be compared with the first.
    output_image: Optional[Image] (= None)
        The resulting binary image where pixels will be 1 only if source1
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_equal
    """
    return clic._equal(device, input_image0, input_image1, output_image)

@plugin_function(categories=["binarize", "in assistant"])
def equal_constant(
    input_image: Image,
    output_image: Optional[Image] =None,
    scalar: float =0,
    device: Optional[Device] =None
) -> Image:
    """Determines if an image A and a constant b are equal. <pre>f(a, b) = 1 if a == b;
    0 otherwise.</pre>

    Parameters
    ----------
    input_image: Image 
        The image where every pixel is compared to the constant.
    output_image: Optional[Image] (= None)
        The resulting binary image where pixels will be 1 only if source1
    scalar: float (= 0)
        The constant where every pixel is compared to.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_equalConstant
    """
    return clic._equal_constant(device, input_image, output_image, float(scalar))

@plugin_function(categories=["binary processing"])
def erode_box(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Computes a binary image with pixel values 0 and 1 containing the binary erosion
    of a given input image. The erosion takes the Mooreneighborhood (8 pixels in 2D
    and 26 pixels in 3d) into account. The pixels in the input image with pixel
    value not equal to 0 will be interpreted as 1. This method is comparable to the
    'Erode' menu in ImageJ in case it is applied to a 2D image. The only difference
    is that the output image contains values 0 and 1 instead of 0 and 255.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_erodeBox
    """
    return clic._erode_box(device, input_image, output_image)

@plugin_function(categories=["binary processing"])
def erode_sphere(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Computes a binary image with pixel values 0 and 1 containing the binary erosion
    of a given input image. The erosion takes the vonNeumannneighborhood (4 pixels
    in 2D and 6 pixels in 3d) into account. The pixels in the input image with pixel
    value not equal to 0 will be interpreted as 1.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_erodeSphere
    """
    return clic._erode_sphere(device, input_image, output_image)

@plugin_function(categories=["binary processing"])
def erode(
    input_image: Image,
    output_image: Optional[Image] =None,
    connectivity: str ="box",
    device: Optional[Device] =None
) -> Image:
    """Computes a binary image with pixel values 0 and 1 containing the binary erosion
    of a given input image. The erosion apply the Mooreneighborhood (8 pixels in 2D
    and 26 pixels in 3d) for the "box" connectivity and the vonNeumannneighborhood
    (4 pixels in 2D and 6 pixels in 3d) for a "sphere" connectivity. The pixels in
    the input image with pixel value not equal to 0 will be interpreted as 1.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    connectivity: str (= "box")
        Element shape, "box" or "sphere".
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_erodeBox
    [2] https://clij.github.io/clij2-docs/reference_erodeSphere
    """
    return clic._erode(device, input_image, output_image, str(connectivity))

@plugin_function(categories=["filter", "in assistant"])
def exponential(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Computes base exponential of all pixels values. f(x) = exp(x) Author(s): Peter
    Haub, Robert Haase

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_exponential
    """
    return clic._exponential(device, input_image, output_image)

@plugin_function
def flip(
    input_image: Image,
    output_image: Optional[Image] =None,
    flip_x: bool =True,
    flip_y: bool =True,
    flip_z: bool =True,
    device: Optional[Device] =None
) -> Image:
    """Flips an image in X, Y and/or Z direction depending on boolean flags.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    flip_x: bool (= True)
        Flip along the x axis if true.
    flip_y: bool (= True)
        Flip along the y axis if true.
    flip_z: bool (= True)
        Flip along the z axis if true.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_flip3D
    """
    return clic._flip(device, input_image, output_image, flip_x, flip_y, flip_z)

@plugin_function(categories=["filter", "denoise", "in assistant", "bia-bob-suggestion"])
def gaussian_blur(
    input_image: Image,
    output_image: Optional[Image] =None,
    sigma_x: float =0,
    sigma_y: float =0,
    sigma_z: float =0,
    device: Optional[Device] =None
) -> Image:
    """Computes the Gaussian blurred image of an image given sigma values in X, Y and
    Z. Thus, the filter kernel can have nonisotropic shape. The implementation is
    done separable. In case a sigma equals zero, the direction is not blurred.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    sigma_x: float (= 0)
        Sigma value along the x axis.
    sigma_y: float (= 0)
        Sigma value along the y axis.
    sigma_z: float (= 0)
        Sigma value along the z axis.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_gaussianBlur3D
    """
    return clic._gaussian_blur(device, input_image, output_image, float(sigma_x), float(sigma_y), float(sigma_z))

@plugin_function
def generate_distance_matrix(
    input_image0: Image,
    input_image1: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Computes the distance between all point coordinates given in two point lists.
    Takes two images containing pointlists (dimensionality n * d, n: number of
    points and d: dimensionality) and builds up a matrix containing the distances
    between these points. Convention: Given two point lists with dimensionality n *
    d and m * d, the distance matrix will be of size(n + 1) * (m + 1). The first row
    and column contain zeros. They represent the distance of the (see
    generateTouchMatrix). Thus, one can threshold a distance matrix to generate a
    touch matrix out of it for drawing objects to a theoretical background object.
    In that way, distance matrices are of the same size as touch matrices meshes.

    Parameters
    ----------
    input_image0: Image 
        First input image to process.
    input_image1: Image 
        Second input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_generateDistanceMatrix
    """
    return clic._generate_distance_matrix(device, input_image0, input_image1, output_image)

@plugin_function(categories=["filter", "edge detection", "in assistant"])
def gradient_x(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Computes the gradient of gray values along X. Assuming a, b and c are three
    adjacent pixels in X direction. In the target image will be saved as: <pre>b' =
    c a;</pre>

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_gradientX
    """
    return clic._gradient_x(device, input_image, output_image)

@plugin_function(categories=["filter", "edge detection", "in assistant"])
def gradient_y(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Computes the gradient of gray values along Y. Assuming a, b and c are three
    adjacent pixels in Y direction. In the target image will be saved as: <pre>b' =
    c a;</pre>

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_gradientY
    """
    return clic._gradient_y(device, input_image, output_image)

@plugin_function(categories=["filter", "edge detection", "in assistant"])
def gradient_z(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Computes the gradient of gray values along Z. Assuming a, b and c are three
    adjacent pixels in Z direction. In the target image will be saved as: <pre>b' =
    c a;</pre>

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_gradientZ
    """
    return clic._gradient_z(device, input_image, output_image)

@plugin_function(categories=["combine", "binarize", "in assistant"])
def greater(
    input_image0: Image,
    input_image1: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Determines if two images A and B greater pixel wise. f(a, b) = 1 if a > b; 0
    otherwise.

    Parameters
    ----------
    input_image0: Image 
        First input image to process.
    input_image1: Image 
        Second input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_greater
    """
    return clic._greater(device, input_image0, input_image1, output_image)

@plugin_function(categories=["binarize", "in assistant"])
def greater_constant(
    input_image: Image,
    output_image: Optional[Image] =None,
    scalar: float =0,
    device: Optional[Device] =None
) -> Image:
    """Determines if two images A and B greater pixel wise. f(a, b) = 1 if a > b; 0
    otherwise.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    scalar: float (= 0)
        
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_greaterConstant
    """
    return clic._greater_constant(device, input_image, output_image, float(scalar))

@plugin_function(categories=["combine", "binarize", "in assistant"])
def greater_or_equal(
    input_image0: Image,
    input_image1: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Determines if two images A and B greater or equal pixel wise. f(a, b) = 1 if a
    >= b; 0 otherwise.

    Parameters
    ----------
    input_image0: Image 
        First input image to process.
    input_image1: Image 
        Second input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_greaterOrEqual
    """
    return clic._greater_or_equal(device, input_image0, input_image1, output_image)

@plugin_function(categories=["binarize", "in assistant"])
def greater_or_equal_constant(
    input_image: Image,
    output_image: Optional[Image] =None,
    scalar: float =0,
    device: Optional[Device] =None
) -> Image:
    """Determines if two images A and B greater or equal pixel wise. f(a, b) = 1 if a
    >= b; 0 otherwise.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    scalar: float (= 0)
        Scalar value used in the comparison.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_greaterOrEqualConstant
    """
    return clic._greater_or_equal_constant(device, input_image, output_image, float(scalar))

@plugin_function
def hessian_eigenvalues(
    input_image: Image,
    small_eigenvalue: Optional[Image] =None,
    middle_eigenvalue: Optional[Image] =None,
    large_eigenvalue: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Computes the eigenvalues of the hessian matrix of a 2d or 3d image. Hessian
    matrix or 2D images: [Ixx, Ixy] [Ixy, Iyy] Hessian matrix for 3D images: [Ixx,
    Ixy, Ixz] [Ixy, Iyy, Iyz] [Ixz, Iyz, Izz] Ixx denotes the second derivative in
    x. Ixx and Iyy are calculated by convolving the image with the 1d kernel [1 2
    1]. Ixy is calculated by a convolution with the 2d kernel: [ 0.25 0 0.25] [ 0 0
    0] [0.25 0 0.25] Note: This is the only clesperanto function that returns
    multiple images. This API might be subject to change in the future. Consider
    using small_hessian_eigenvalue() and/or large_hessian_eigenvalue() instead which
    return only one image.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    small_eigenvalue: Optional[Image] (= None)
        Output result image.
    middle_eigenvalue: Optional[Image] (= None)
        Output result image, null if input is 2D.
    large_eigenvalue: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._hessian_eigenvalues(device, input_image, small_eigenvalue, middle_eigenvalue, large_eigenvalue)

@plugin_function(categories=["filter", "edge detection", "in assistant"])
def laplace_box(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Applies the Laplace operator (Box neighborhood) to an image.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_laplaceBox
    """
    return clic._laplace_box(device, input_image, output_image)

@plugin_function(categories=["filter", "edge detection"])
def laplace_diamond(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Applies the Laplace operator (Diamond neighborhood) to an image.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_laplaceDiamond
    """
    return clic._laplace_diamond(device, input_image, output_image)

@plugin_function(categories=["filter", "edge detection"])
def laplace(
    input_image: Image,
    output_image: Optional[Image] =None,
    connectivity: str ="box",
    device: Optional[Device] =None
) -> Image:
    """Applies the Laplace operator with a "box" or a "sphere" neighborhood to an
    image.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    connectivity: str (= "box")
        Filter neigborhood
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_laplaceDiamond
    """
    return clic._laplace(device, input_image, output_image, str(connectivity))

@plugin_function(categories=["filter", "combine", "in assistant"])
def local_cross_correlation(
    input_image0: Image,
    input_image1: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Compute the cross correlation of an image to a given kernel.

    Parameters
    ----------
    input_image0: Image 
        First input image to process.
    input_image1: Image 
        Second input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._local_cross_correlation(device, input_image0, input_image1, output_image)

@plugin_function(categories=["filter", "in assistant"])
def logarithm(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Computes base e logarithm of all pixels values. f(x) = log(x) Author(s): Peter
    Haub, Robert Haase

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_logarithm
    """
    return clic._logarithm(device, input_image, output_image)

@plugin_function
def mask(
    input_image: Image,
    mask: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Computes a masked image by applying a binary mask to an image. All pixel values
    x of image X will be copied to the destination image in case pixel value m at
    the same position in the mask image is not equal to zero. <pre>f(x,m) = (x if (m
    != 0); (0 otherwise))</pre>

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    mask: Image 
        Mask image to apply.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_mask
    """
    return clic._mask(device, input_image, mask, output_image)

@plugin_function
def mask_label(
    input_image0: Image,
    input_image1: Image,
    output_image: Optional[Image] =None,
    label: float =1,
    device: Optional[Device] =None
) -> Image:
    """Computes a masked image by applying a label mask to an image. All pixel values x
    of image X will be copied to the destination image in case pixel value m at the
    same position in the label_map image has the right index value i. f(x,m,i) = (x
    if (m == i); (0 otherwise))

    Parameters
    ----------
    input_image0: Image 
        Input Intensity image.
    input_image1: Image 
        Input Label image.
    output_image: Optional[Image] (= None)
        Output result image.
    label: float (= 1)
        Label value to use.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_maskLabel
    """
    return clic._mask_label(device, input_image0, input_image1, output_image, float(label))

@plugin_function(categories=["filter", "in assistant"])
def maximum_image_and_scalar(
    input_image: Image,
    output_image: Optional[Image] =None,
    scalar: float =0,
    device: Optional[Device] =None
) -> Image:
    """Computes the maximum of a constant scalar s and each pixel value x in a given
    image X. <pre>f(x, s) = max(x, s)</pre>

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    scalar: float (= 0)
        Scalar value used in the comparison.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_maximumImageAndScalar
    """
    return clic._maximum_image_and_scalar(device, input_image, output_image, float(scalar))

@plugin_function(categories=["combine", "in assistant"])
def maximum_images(
    input_image0: Image,
    input_image1: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Computes the maximum of a pair of pixel values x, y from two given images X and
    Y. <pre>f(x, y) = max(x, y)</pre>

    Parameters
    ----------
    input_image0: Image 
        First input image to process.
    input_image1: Image 
        Second input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_maximumImages
    """
    return clic._maximum_images(device, input_image0, input_image1, output_image)

@plugin_function(categories=["filter", "in assistant"])
def maximum_box(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: int =1,
    radius_y: int =1,
    radius_z: int =1,
    device: Optional[Device] =None
) -> Image:
    """Computes the local maximum of a pixels cube neighborhood. The cubes size is
    specified by its halfwidth, halfheight and halfdepth (radius).

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    radius_x: int (= 1)
        Radius size along x axis.
    radius_y: int (= 1)
        Radius size along y axis.
    radius_z: int (= 1)
        Radius size along z axis.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_maximum3DBox
    """
    return clic._maximum_box(device, input_image, output_image, int(radius_x), int(radius_y), int(radius_z))

@plugin_function(categories=["filter", "in assistant"])
def maximum(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: int =0,
    radius_y: int =0,
    radius_z: int =0,
    connectivity: str ="box",
    device: Optional[Device] =None
) -> Image:
    """Computes the local maximum of a pixels neighborhood (box or sphere). The
    neighborhood size is specified by its halfwidth, halfheight and halfdepth
    (radius).

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    radius_x: int (= 0)
        Radius size along x axis.
    radius_y: int (= 0)
        Radius size along y axis.
    radius_z: int (= 0)
        Radius size along z axis.
    connectivity: str (= "box")
        Filter neigborhood
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_maximum3DBox
    [2] https://clij.github.io/clij2-docs/reference_maximum3DSphere
    """
    return clic._maximum(device, input_image, output_image, int(radius_x), int(radius_y), int(radius_z), str(connectivity))

@plugin_function(categories=["projection"])
def maximum_x_projection(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Determines the maximum intensity projection of an image along X.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_maximumXProjection
    """
    return clic._maximum_x_projection(device, input_image, output_image)

@plugin_function(categories=["projection"])
def maximum_y_projection(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Determines the maximum intensity projection of an image along X.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_maximumYProjection
    """
    return clic._maximum_y_projection(device, input_image, output_image)

@plugin_function(categories=["projection", "in assistant", "bia-bob-suggestion"])
def maximum_z_projection(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Determines the maximum intensity projection of an image along Z.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_maximumZProjection
    """
    return clic._maximum_z_projection(device, input_image, output_image)

@plugin_function(categories=["filter", "denoise", "in assistant"])
def mean_box(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: int =1,
    radius_y: int =1,
    radius_z: int =1,
    device: Optional[Device] =None
) -> Image:
    """Computes the local mean average of a pixels boxshaped neighborhood. The cubes
    size is specified by its halfwidth, halfheight and halfdepth (radius).

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    radius_x: int (= 1)
        Radius size along x axis.
    radius_y: int (= 1)
        Radius size along y axis.
    radius_z: int (= 1)
        Radius size along z axis.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_mean3DBox
    """
    return clic._mean_box(device, input_image, output_image, int(radius_x), int(radius_y), int(radius_z))

@plugin_function(categories=["filter", "denoise", "in assistant", "bia-bob-suggestion"])
def mean_sphere(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: int =1,
    radius_y: int =1,
    radius_z: int =1,
    device: Optional[Device] =None
) -> Image:
    """Computes the local mean average of a pixels spherical neighborhood. The spheres
    size is specified by its halfwidth, halfheight and halfdepth (radius).

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    radius_x: int (= 1)
        Radius size along x axis.
    radius_y: int (= 1)
        Radius size along y axis.
    radius_z: int (= 1)
        Radius size along z axis.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_mean3DSphere
    """
    return clic._mean_sphere(device, input_image, output_image, int(radius_x), int(radius_y), int(radius_z))

@plugin_function(categories=["filter", "denoise", "in assistant"])
def mean(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: int =1,
    radius_y: int =1,
    radius_z: int =1,
    connectivity: str ="box",
    device: Optional[Device] =None
) -> Image:
    """Computes the local mean average of a pixels neighborhood defined as a boxshaped
    or a sphereshaped. The shape size is specified by its halfwidth, halfheight and
    halfdepth (radius).

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    radius_x: int (= 1)
        Radius size along x axis.
    radius_y: int (= 1)
        Radius size along y axis.
    radius_z: int (= 1)
        Radius size along z axis.
    connectivity: str (= "box")
        Filter neigborhood
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_mean3DSphere
    """
    return clic._mean(device, input_image, output_image, int(radius_x), int(radius_y), int(radius_z), str(connectivity))

@plugin_function(categories=["projection"])
def mean_x_projection(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Determines the mean average intensity projection of an image along X.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_meanXProjection
    """
    return clic._mean_x_projection(device, input_image, output_image)

@plugin_function(categories=["projection"])
def mean_y_projection(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Determines the mean average intensity projection of an image along Y.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_meanYProjection
    """
    return clic._mean_y_projection(device, input_image, output_image)

@plugin_function(categories=["projection", "in assistant", "bia-bob-suggestion"])
def mean_z_projection(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Determines the mean average intensity projection of an image along Z.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_meanZProjection
    """
    return clic._mean_z_projection(device, input_image, output_image)

@plugin_function(categories=["filter", "denoise", "in assistant"])
def median_box(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: int =1,
    radius_y: int =1,
    radius_z: int =1,
    device: Optional[Device] =None
) -> Image:
    """Computes the local median of a pixels box shaped neighborhood. The box is
    specified by its halfwidth and halfheight (radius). For technical reasons, the
    area of the box must have less than 1000 pixels.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    radius_x: int (= 1)
        Radius size along x axis.
    radius_y: int (= 1)
        Radius size along y axis.
    radius_z: int (= 1)
        Radius size along z axis.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_median3DBox
    """
    return clic._median_box(device, input_image, output_image, int(radius_x), int(radius_y), int(radius_z))

@plugin_function(categories=["filter", "denoise", "in assistant"])
def median_sphere(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: int =1,
    radius_y: int =1,
    radius_z: int =1,
    device: Optional[Device] =None
) -> Image:
    """Computes the local median of a pixels sphere shaped neighborhood. The sphere is
    specified by its halfwidth and halfheight (radius). For technical reasons, the
    area of the box must have less than 1000 pixels.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    radius_x: int (= 1)
        Radius size along x axis.
    radius_y: int (= 1)
        Radius size along y axis.
    radius_z: int (= 1)
        Radius size along z axis.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_median3DSphere
    """
    return clic._median_sphere(device, input_image, output_image, int(radius_x), int(radius_y), int(radius_z))

@plugin_function(categories=["filter", "denoise", "in assistant"])
def median(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: int =1,
    radius_y: int =1,
    radius_z: int =1,
    connectivity: str ="box",
    device: Optional[Device] =None
) -> Image:
    """Computes the local median of a pixels neighborhood. The neighborhood is defined
    as a box or a sphere shape. Its size is specified by its halfwidth, halfheight,
    and halfdepth (radius). For technical reasons, the area of the shpae must have
    less than 1000 pixels.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    radius_x: int (= 1)
        Radius size along x axis.
    radius_y: int (= 1)
        Radius size along y axis.
    radius_z: int (= 1)
        Radius size along z axis.
    connectivity: str (= "box")
        Filter neigborhood
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_median3DSphere
    """
    return clic._median(device, input_image, output_image, int(radius_x), int(radius_y), int(radius_z), str(connectivity))

@plugin_function(categories=["filter", "in assistant"])
def minimum_box(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: int =0,
    radius_y: int =0,
    radius_z: int =0,
    device: Optional[Device] =None
) -> Image:
    """Computes the local minimum of a pixels cube neighborhood. The cubes size is
    specified by its halfwidth, halfheight and halfdepth (radius).

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    radius_x: int (= 0)
        Radius size along x axis.
    radius_y: int (= 0)
        Radius size along y axis.
    radius_z: int (= 0)
        Radius size along z axis.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_minimum3DBox
    """
    return clic._minimum_box(device, input_image, output_image, int(radius_x), int(radius_y), int(radius_z))

@plugin_function(categories=["filter", "in assistant"])
def minimum(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: int =0,
    radius_y: int =0,
    radius_z: int =0,
    connectivity: str ="box",
    device: Optional[Device] =None
) -> Image:
    """Computes the local minimum of a pixels cube neighborhood. The cubes size is
    specified by its halfwidth, halfheight and halfdepth (radius).

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    radius_x: int (= 0)
        Radius size along x axis.
    radius_y: int (= 0)
        Radius size along y axis.
    radius_z: int (= 0)
        Radius size along z axis.
    connectivity: str (= "box")
        Filter neigborhood
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_minimum3DBox
    [2] https://clij.github.io/clij2-docs/reference_minimum3DSphere
    """
    return clic._minimum(device, input_image, output_image, int(radius_x), int(radius_y), int(radius_z), str(connectivity))

@plugin_function(categories=["filter", "in assistant"])
def minimum_image_and_scalar(
    input_image: Image,
    output_image: Optional[Image] =None,
    scalar: float =0,
    device: Optional[Device] =None
) -> Image:
    """Computes the minimum of a constant scalar s and each pixel value x in a given
    image X. <pre>f(x, s) = min(x, s)</pre>

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    scalar: float (= 0)
        Scalar value used in the comparison.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_minimumImageAndScalar
    """
    return clic._minimum_image_and_scalar(device, input_image, output_image, float(scalar))

@plugin_function(categories=["combine", "in assistant"])
def minimum_images(
    input_image0: Image,
    input_image1: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Computes the minimum of a pair of pixel values x, y from two given images X and
    Y. <pre>f(x, y) = min(x, y)</pre>

    Parameters
    ----------
    input_image0: Image 
        First input image to process.
    input_image1: Image 
        Second input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_minimumImages
    """
    return clic._minimum_images(device, input_image0, input_image1, output_image)

@plugin_function(categories=["projection"])
def minimum_x_projection(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Determines the minimum intensity projection of an image along Y.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_minimumXProjection
    """
    return clic._minimum_x_projection(device, input_image, output_image)

@plugin_function(categories=["projection"])
def minimum_y_projection(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Determines the minimum intensity projection of an image along Y.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_minimumYProjection
    """
    return clic._minimum_y_projection(device, input_image, output_image)

@plugin_function(categories=["projection", "in assistant", "bia-bob-suggestion"])
def minimum_z_projection(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Determines the minimum intensity projection of an image along Z.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_minimumZProjection
    """
    return clic._minimum_z_projection(device, input_image, output_image)

@plugin_function(categories=["label processing", "in assistant"])
def mode_box(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: int =1,
    radius_y: int =1,
    radius_z: int =1,
    device: Optional[Device] =None
) -> Image:
    """Computes the local mode of a pixels box shaped neighborhood. This can be used to
    postprocess and locally correct semantic segmentation results. The box is
    specified by its halfwidth and halfheight (radius). For technical reasons, the
    intensities must lie within a range from 0 to 255. In case multiple values have
    maximum frequency, the smallest one is returned.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    radius_x: int (= 1)
        Radius size along x axis.
    radius_y: int (= 1)
        Radius size along y axis.
    radius_z: int (= 1)
        Radius size along z axis.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._mode_box(device, input_image, output_image, int(radius_x), int(radius_y), int(radius_z))

@plugin_function(categories=["label processing", "in assistant", "bia-bob-suggestion"])
def mode_sphere(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: int =1,
    radius_y: int =1,
    radius_z: int =1,
    device: Optional[Device] =None
) -> Image:
    """Computes the local mode of a pixels sphere shaped neighborhood. This can be used
    to postprocess and locally correct semantic segmentation results. The sphere is
    specified by its halfwidth and halfheight (radius). For technical reasons, the
    intensities must lie within a range from 0 to 255. In case multiple values have
    maximum frequency, the smallest one is returned.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    radius_x: int (= 1)
        Radius size along x axis.
    radius_y: int (= 1)
        Radius size along y axis.
    radius_z: int (= 1)
        Radius size along z axis.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._mode_sphere(device, input_image, output_image, int(radius_x), int(radius_y), int(radius_z))

@plugin_function(categories=["label processing", "in assistant"])
def mode(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: int =1,
    radius_y: int =1,
    radius_z: int =1,
    connectivity: str ="box",
    device: Optional[Device] =None
) -> Image:
    """Computes the local mode of a pixels neighborhood. This neighborhood can be
    shaped as a box or a sphere. This can be used to postprocess and locally correct
    semantic segmentation results. The shape size is specified by its halfwidth,
    halfheight, and halfdepth (radius). For technical reasons, the intensities must
    lie within a range from 0 to 255 (uint8). In case multiple values have maximum
    frequency, the smallest one is returned.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    radius_x: int (= 1)
        Radius size along x axis.
    radius_y: int (= 1)
        Radius size along y axis.
    radius_z: int (= 1)
        Radius size along z axis.
    connectivity: str (= "box")
        Filter neigborhood
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._mode(device, input_image, output_image, int(radius_x), int(radius_y), int(radius_z), str(connectivity))

@plugin_function(categories=["combine"])
def modulo_images(
    input_image0: Image,
    input_image1: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Computes the remainder of a division of pairwise pixel values in two images

    Parameters
    ----------
    input_image0: Image 
        First input image to process.
    input_image1: Image 
        Second input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._modulo_images(device, input_image0, input_image1, output_image)

@plugin_function
def multiply_image_and_position(
    input_image: Image,
    output_image: Optional[Image] =None,
    dimension: int =0,
    device: Optional[Device] =None
) -> Image:
    """Multiplies all pixel intensities with the x, y or z coordinate, depending on
    specified dimension.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    dimension: int (= 0)
        Dimension (0,1,2) to use in the operation.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_multiplyImageAndCoordinate
    """
    return clic._multiply_image_and_position(device, input_image, output_image, int(dimension))

@plugin_function(categories=["filter", "in assistant"])
def multiply_image_and_scalar(
    input_image: Image,
    output_image: Optional[Image] =None,
    scalar: float =0,
    device: Optional[Device] =None
) -> Image:
    """Multiplies all pixels value x in a given image X with a constant scalar s.
    <pre>f(x, s) = x * s</pre>

    Parameters
    ----------
    input_image: Image 
        The input image to be multiplied with a constant.
    output_image: Optional[Image] (= None)
        The output image where results are written into.
    scalar: float (= 0)
        The number with which every pixel will be multiplied with.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_multiplyImageAndScalar
    """
    return clic._multiply_image_and_scalar(device, input_image, output_image, float(scalar))

@plugin_function(categories=["combine", "in assistant"])
def multiply_images(
    input_image0: Image,
    input_image1: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Multiplies all pairs of pixel values x and y from two image X and Y. <pre>f(x,
    y) = x * y</pre>

    Parameters
    ----------
    input_image0: Image 
        The first input image to be multiplied.
    input_image1: Image 
        The second image to be multiplied.
    output_image: Optional[Image] (= None)
        The output image where results are written into.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_multiplyImages
    """
    return clic._multiply_images(device, input_image0, input_image1, output_image)

@plugin_function
def nan_to_num(
    input_image: Image,
    output_image: Optional[Image] =None,
    nan: float =0,
    posinf: float =np.nan_to_num(float('inf')),
    neginf: float =np.nan_to_num(float('-inf')),
    device: Optional[Device] =None
) -> Image:
    """Copies all pixels instead those which are not a number (NaN), or
    positive/negative infinity which are replaced by a defined new value, default 0.
    This function aims to work similarly as its counterpart in numpy [1]. Default
    values for posinf and neginf may differ from numpy and even differ depending on
    compute hardware. It is recommended to specify those values.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        The output image where results are written into.
    nan: float (= 0)
        Value to replace
    posinf: float (= np.nan_to_num(float('inf')))
        Value to replace +inf with.
    neginf: float (= np.nan_to_num(float('-inf')))
        Value to replace -inf with.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://numpy.org/doc/stable/reference/generated/numpy.nan_to_num.html
    """
    return clic._nan_to_num(device, input_image, output_image, float(nan), float(posinf), float(neginf))

@plugin_function
def nonzero_maximum_box(
    input_image: Image,
    output_image0: Image,
    output_image1: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Apply a maximum filter (box shape) to the input image. The radius is fixed to 1
    and pixels with value 0 are ignored. Note: Pixels with 0 value in the input
    image will not be overwritten in the output image. Thus, the result image should
    be initialized by copying the original image in advance.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image0: Image 
        Output flag (0 or 1).
    output_image1: Optional[Image] (= None)
        Output image where results are written into.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_nonzeroMaximumBox
    """
    return clic._nonzero_maximum_box(device, input_image, output_image0, output_image1)

@plugin_function
def nonzero_maximum_diamond(
    input_image: Image,
    output_image0: Image,
    output_image1: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Apply a maximum filter (diamond shape) to the input image. The radius is fixed
    to 1 and pixels with value 0 are ignored. Note: Pixels with 0 value in the input
    image will not be overwritten in the output image. Thus, the result image should
    be initialized by copying the original image in advance.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image0: Image 
        Output flag (0 or 1).
    output_image1: Optional[Image] (= None)
        Output image where results are written into.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_nonzeroMaximumDiamond
    """
    return clic._nonzero_maximum_diamond(device, input_image, output_image0, output_image1)

@plugin_function
def nonzero_maximum(
    input_image: Image,
    output_image0: Image,
    output_image1: Optional[Image] =None,
    connectivity: str ="box",
    device: Optional[Device] =None
) -> Image:
    """Apply a maximum filter of a neighborhood to the input image. The neighborhood
    shape can be a box or a sphere. The size is fixed to 1 and pixels with value 0
    are ignored. Note: Pixels with 0 value in the input image will not be
    overwritten in the output image. Thus, the result image should be initialized by
    copying the original image in advance.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image0: Image 
        Output flag (0 or 1).
    output_image1: Optional[Image] (= None)
        Output image where results are written into.
    connectivity: str (= "box")
        Filter neigborhood
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_nonzeroMaximumBox
    [2] https://clij.github.io/clij2-docs/reference_nonzeroMaximumDiamond
    """
    return clic._nonzero_maximum(device, input_image, output_image0, output_image1, str(connectivity))

@plugin_function
def nonzero_minimum_box(
    input_image: Image,
    output_image0: Image,
    output_image1: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Apply a minimum filter (box shape) to the input image. The radius is fixed to 1
    and pixels with value 0 are ignored. Note: Pixels with 0 value in the input
    image will not be overwritten in the output image. Thus, the result image should
    be initialized by copying the original image in advance.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image0: Image 
        Output flag (0 or 1).
    output_image1: Optional[Image] (= None)
        Output image where results are written into.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_nonzeroMinimumBox
    """
    return clic._nonzero_minimum_box(device, input_image, output_image0, output_image1)

@plugin_function
def nonzero_minimum_diamond(
    input_image: Image,
    output_image0: Image,
    output_image1: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Apply a minimum filter (diamond shape) to the input image. The radius is fixed
    to 1 and pixels with value 0 are ignored.Note: Pixels with 0 value in the input
    image will not be overwritten in the output image. Thus, the result image should
    be initialized by copying the original image in advance.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image0: Image 
        Output flag (0 or 1).
    output_image1: Optional[Image] (= None)
        Output image where results are written into.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_nonzeroMinimumDiamond
    """
    return clic._nonzero_minimum_diamond(device, input_image, output_image0, output_image1)

@plugin_function
def nonzero_minimum(
    input_image: Image,
    output_image0: Image,
    output_image1: Optional[Image] =None,
    connectivity: str ="box",
    device: Optional[Device] =None
) -> Image:
    """Apply a minimum filter of a neighborhood to the input image. The neighborhood
    shape can be a box or a sphere. The radius is fixed to 1 and pixels with value 0
    are ignored.Note: Pixels with 0 value in the input image will not be overwritten
    in the output image. Thus, the result image should be initialized by copying the
    original image in advance.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image0: Image 
        Output flag (0 or 1).
    output_image1: Optional[Image] (= None)
        Output image where results are written into.
    connectivity: str (= "box")
        Filter neigborhood
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_nonzeroMinimumBox
    [2] https://clij.github.io/clij2-docs/reference_nonzeroMinimumDiamond
    """
    return clic._nonzero_minimum(device, input_image, output_image0, output_image1, str(connectivity))

@plugin_function(categories=["combine", "binarize", "in assistant"])
def not_equal(
    input_image0: Image,
    input_image1: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Determines if two images A and B equal pixel wise. f(a, b) = 1 if a != b; 0
    otherwise.

    Parameters
    ----------
    input_image0: Image 
        The first image to be compared with.
    input_image1: Image 
        The second image to be compared with the first.
    output_image: Optional[Image] (= None)
        The resulting binary image where pixels will be 1 only if source1
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_notEqual
    """
    return clic._not_equal(device, input_image0, input_image1, output_image)

@plugin_function(categories=["binarize", "in assistant"])
def not_equal_constant(
    input_image: Image,
    output_image: Optional[Image] =None,
    scalar: float =0,
    device: Optional[Device] =None
) -> Image:
    """Determines if two images A and B equal pixel wise. f(a, b) = 1 if a != b; 0
    otherwise.

    Parameters
    ----------
    input_image: Image 
        The image where every pixel is compared to the constant.
    output_image: Optional[Image] (= None)
        The resulting binary image where pixels will be 1 only if source1
    scalar: float (= 0)
        The constant where every pixel is compared to.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_notEqualConstant
    """
    return clic._not_equal_constant(device, input_image, output_image, float(scalar))

@plugin_function(categories=["combine", "in assistant"])
def paste(
    input_image: Image,
    output_image: Optional[Image] =None,
    index_x: int =0,
    index_y: int =0,
    index_z: int =0,
    device: Optional[Device] =None
) -> Image:
    """Pastes an image into another image at a given position.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    index_x: int (= 0)
        Origin pixel coodinate in x to paste.
    index_y: int (= 0)
        Origin pixel coodinate in y to paste.
    index_z: int (= 0)
        Origin pixel coodinate in z to paste.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_paste3D
    """
    return clic._paste(device, input_image, output_image, int(index_x), int(index_y), int(index_z))

@plugin_function
def onlyzero_overwrite_maximum_box(
    input_image: Image,
    output_image0: Image,
    output_image1: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Apply a local maximum filter to an image which only overwrites pixels with value
    0.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image0: Image 
        Output flag value, 0 or 1.
    output_image1: Optional[Image] (= None)
        Output image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_onlyzeroOverwriteMaximumBox
    """
    return clic._onlyzero_overwrite_maximum_box(device, input_image, output_image0, output_image1)

@plugin_function
def onlyzero_overwrite_maximum_diamond(
    input_image: Image,
    output_image0: Image,
    output_image1: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Apply a local maximum filter to an image which only overwrites pixels with value
    0.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image0: Image 
        Output flag value, 0 or 1.
    output_image1: Optional[Image] (= None)
        Output image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_onlyzeroOverwriteMaximumDiamond
    """
    return clic._onlyzero_overwrite_maximum_diamond(device, input_image, output_image0, output_image1)

@plugin_function
def onlyzero_overwrite_maximum(
    input_image: Image,
    output_image0: Image,
    output_image1: Optional[Image] =None,
    connectivity: str ="box",
    device: Optional[Device] =None
) -> Image:
    """Apply a local maximum filter to an image which only overwrites pixels with value
    0.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image0: Image 
        Output flag value, 0 or 1.
    output_image1: Optional[Image] (= None)
        Output image.
    connectivity: str (= "box")
        Filter neigborhood
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_onlyzeroOverwriteMaximumBox
    [2] https://clij.github.io/clij2-docs/reference_onlyzeroOverwriteMaximumDiamond
    """
    return clic._onlyzero_overwrite_maximum(device, input_image, output_image0, output_image1, str(connectivity))

@plugin_function(categories=["filter", "in assistant"])
def power(
    input_image: Image,
    output_image: Optional[Image] =None,
    scalar: float =1,
    device: Optional[Device] =None
) -> Image:
    """Computes all pixels value x to the power of a given exponent a. <pre>f(x, a) = x
    ^ a</pre>

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    scalar: float (= 1)
        Power value.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_power
    """
    return clic._power(device, input_image, output_image, float(scalar))

@plugin_function(categories=["combine", "in assistant"])
def power_images(
    input_image0: Image,
    input_image1: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Calculates x to the power of y pixel wise of two images X and Y.

    Parameters
    ----------
    input_image0: Image 
        First input image to process.
    input_image1: Image 
        Second input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_powerImages
    """
    return clic._power_images(device, input_image0, input_image1, output_image)

@plugin_function(categories=["transform", "in assistant"])
def range(
    input_image: Image,
    output_image: Optional[Image] =None,
    start_x: Optional[int] =None,
    stop_x: Optional[int] =None,
    step_x: Optional[int] =None,
    start_y: Optional[int] =None,
    stop_y: Optional[int] =None,
    step_y: Optional[int] =None,
    start_z: Optional[int] =None,
    stop_z: Optional[int] =None,
    step_z: Optional[int] =None,
    device: Optional[Device] =None
) -> Image:
    """Crops an image according to a defined range and step size.

    Parameters
    ----------
    input_image: Image 
        First input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    start_x: Optional[int] (= None)
        Range starting value in x
    stop_x: Optional[int] (= None)
        Range stop value in x
    step_x: Optional[int] (= None)
        Range step value in x
    start_y: Optional[int] (= None)
        Range starting value in y
    stop_y: Optional[int] (= None)
        Range stop value in y
    step_y: Optional[int] (= None)
        Range step value in y
    start_z: Optional[int] (= None)
        Range starting value in z
    stop_z: Optional[int] (= None)
        Range stop value in z
    step_z: Optional[int] (= None)
        Range step value in z
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._range(device, input_image, output_image, start_x, stop_x, step_x, start_y, stop_y, step_y, start_z, stop_z, step_z)

@plugin_function
def read_values_from_positions(
    input_image: Image,
    list: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Go to positions in a given image specified by a pointlist and read intensities
    of those pixels. The intensities are stored in a new vector. The positions are
    passed as a (x,y,z) coordinate per column.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    list: Image 
        List of coordinate, as a 2D matrix.
    output_image: Optional[Image] (= None)
        Output vector image of intensities.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._read_values_from_positions(device, input_image, list, output_image)

@plugin_function(categories=["bia-bob-suggestion"])
def replace_values(
    input_image0: Image,
    input_image1: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Replaces integer intensities specified in a vector image. The values are passed
    as a vector of values. The vector index represents the old intensity and the
    value at that position represents the new intensity.s

    Parameters
    ----------
    input_image0: Image 
        Input image to process.
    input_image1: Image 
        List of intensities to replace, as a vector of values.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_replaceIntensities
    """
    return clic._replace_values(device, input_image0, input_image1, output_image)

@plugin_function
def replace_value(
    input_image: Image,
    output_image: Optional[Image] =None,
    scalar0: float =0,
    scalar1: float =1,
    device: Optional[Device] =None
) -> Image:
    """Replaces a specific intensity in an image with a given new value.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    scalar0: float (= 0)
        Old value.
    scalar1: float (= 1)
        New value.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_replaceIntensity
    """
    return clic._replace_value(device, input_image, output_image, float(scalar0), float(scalar1))

@plugin_function
def replace_intensity(
    input_image: Image,
    output_image: Optional[Image] =None,
    scalar0: float =0,
    scalar1: float =1,
    device: Optional[Device] =None
) -> Image:
    """Replaces a specific intensity in an image with a given new value.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    scalar0: float (= 0)
        Old value.
    scalar1: float (= 1)
        New value.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_replaceIntensity
    """
    return clic._replace_intensity(device, input_image, output_image, float(scalar0), float(scalar1))

@plugin_function
def replace_intensities(
    input_image0: Image,
    input_image1: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Replaces integer intensities specified in a vector image. The values are passed
    as a vector of values. The vector index represents the old intensity and the
    value at that position represents the new intensity.s

    Parameters
    ----------
    input_image0: Image 
        Input image to process.
    input_image1: Image 
        List of intensities to replace, as a vector of values.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_replaceIntensities
    """
    return clic._replace_intensities(device, input_image0, input_image1, output_image)

@plugin_function(categories=["filter", "in assistant", "bia-bob-suggestion"])
def maximum_sphere(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: float =1,
    radius_y: float =1,
    radius_z: float =0,
    device: Optional[Device] =None
) -> Image:
    """Computes the local maximum of a pixels spherical neighborhood. The spheres size
    is specified by its halfwidth, halfheight and halfdepth (radius).

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    radius_x: float (= 1)
        Radius size along x axis.
    radius_y: float (= 1)
        Radius size along y axis.
    radius_z: float (= 0)
        Radius size along z axis.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_maximum3DSphere
    """
    return clic._maximum_sphere(device, input_image, output_image, float(radius_x), float(radius_y), float(radius_z))

@plugin_function(categories=["filter", "in assistant", "bia-bob-suggestion"])
def minimum_sphere(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: float =1,
    radius_y: float =1,
    radius_z: float =1,
    device: Optional[Device] =None
) -> Image:
    """Computes the local minimum of a pixels spherical neighborhood. The spheres size
    is specified by its halfwidth, halfheight and halfdepth (radius).

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    radius_x: float (= 1)
        Radius size along x axis.
    radius_y: float (= 1)
        Radius size along y axis.
    radius_z: float (= 1)
        Radius size along z axis.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_minimum3DSphere
    """
    return clic._minimum_sphere(device, input_image, output_image, float(radius_x), float(radius_y), float(radius_z))

@plugin_function
def multiply_matrix(
    input_image0: Image,
    input_image1: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Multiplies two matrices with each other. Shape of matrix1 should be equal to
    shape of matrix2 transposed.

    Parameters
    ----------
    input_image0: Image 
        First input image to process.
    input_image1: Image 
        Second input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_multiplyMatrix
    """
    return clic._multiply_matrix(device, input_image0, input_image1, output_image)

@plugin_function(categories=["filter", "in assistant"])
def reciprocal(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Computes 1/x for every pixel value This function is supposed to work similarly
    to its counter part in numpy [1]

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://numpy.org/doc/stable/reference/generated/numpy.reciprocal.html
    """
    return clic._reciprocal(device, input_image, output_image)

@plugin_function
def set(
    input_image: Image,
    scalar: float =0,
    device: Optional[Device] =None
) -> Image:
    """Sets all pixel values x of a given image X to a constant value v. <pre>f(x) =
    v</pre>

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    scalar: float (= 0)
        Value to set.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_set
    """
    return clic._set(device, input_image, float(scalar))

@plugin_function
def set_column(
    input_image: Image,
    column: int =0,
    value: float =0,
    device: Optional[Device] =None
) -> Image:
    """Sets all pixel values x of a given column in X to a constant value v.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    column: int (= 0)
        Column index.
    value: float (= 0)
        Value to set.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_setColumn
    """
    return clic._set_column(device, input_image, int(column), float(value))

@plugin_function
def set_image_borders(
    input_image: Image,
    value: float =0,
    device: Optional[Device] =None
) -> Image:
    """Sets all pixel values at the image border to a given value.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    value: float (= 0)
        Value to set.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_setImageBorders
    """
    return clic._set_image_borders(device, input_image, float(value))

@plugin_function
def set_plane(
    input_image: Image,
    plane: int =0,
    value: float =0,
    device: Optional[Device] =None
) -> Image:
    """Sets all pixel values x of a given plane in X to a constant value v.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    plane: int (= 0)
        Plane index.
    value: float (= 0)
        Value to set.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_setPlane
    """
    return clic._set_plane(device, input_image, int(plane), float(value))

@plugin_function
def set_ramp_x(
    input_image: Image,
    device: Optional[Device] =None
) -> Image:
    """Sets all pixel values to their X coordinate.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_setRampX
    """
    return clic._set_ramp_x(device, input_image)

@plugin_function
def set_ramp_y(
    input_image: Image,
    device: Optional[Device] =None
) -> Image:
    """Sets all pixel values to their Y coordinate.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_setRampY
    """
    return clic._set_ramp_y(device, input_image)

@plugin_function
def set_ramp_z(
    input_image: Image,
    device: Optional[Device] =None
) -> Image:
    """Sets all pixel values to their Z coordinate.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_setRampZ
    """
    return clic._set_ramp_z(device, input_image)

@plugin_function
def set_row(
    input_image: Image,
    row: int =0,
    value: float =0,
    device: Optional[Device] =None
) -> Image:
    """Sets all pixel values x of a given row in X to a constant value v.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    row: int (= 0)
        
    value: float (= 0)
        
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_setRow
    """
    return clic._set_row(device, input_image, int(row), float(value))

@plugin_function
def set_nonzero_pixels_to_pixelindex(
    input_image: Image,
    output_image: Optional[Image] =None,
    offset: int =1,
    device: Optional[Device] =None
) -> Image:
    """Replaces all 0 value pixels in an image with the index of a pixel.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output image.
    offset: int (= 1)
        Offset value to start the indexing.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._set_nonzero_pixels_to_pixelindex(device, input_image, output_image, int(offset))

@plugin_function
def set_where_x_equals_y(
    input_image: Image,
    value: float =0,
    device: Optional[Device] =None
) -> Image:
    """Sets all pixel values a of a given image A to a constant value v in case its
    coordinates x == y. Otherwise the pixel is not overwritten. If you want to
    initialize an identity transfrom matrix, set all pixels to 0 first.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    value: float (= 0)
        Value to set.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_setWhereXequalsY
    """
    return clic._set_where_x_equals_y(device, input_image, float(value))

@plugin_function
def set_where_x_greater_than_y(
    input_image: Image,
    value: float =0,
    device: Optional[Device] =None
) -> Image:
    """Sets all pixel values a of a given image A to a constant value v in case its
    coordinates x > y. Otherwise the pixel is not overwritten. If you want to
    initialize an identity transfrom matrix, set all pixels to 0 first.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    value: float (= 0)
        Value to set.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_setWhereXgreaterThanY
    """
    return clic._set_where_x_greater_than_y(device, input_image, float(value))

@plugin_function
def set_where_x_smaller_than_y(
    input_image: Image,
    value: float =0,
    device: Optional[Device] =None
) -> Image:
    """Sets all pixel values a of a given image A to a constant value v in case its
    coordinates x < y. Otherwise the pixel is not overwritten. If you want to
    initialize an identity transfrom matrix, set all pixels to 0 first.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    value: float (= 0)
        Value to set.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_setWhereXsmallerThanY
    """
    return clic._set_where_x_smaller_than_y(device, input_image, float(value))

@plugin_function
def sign(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Extracts the sign of pixels. If a pixel value < 0, resulting pixel value will be
    1. If it was > 0, it will be 1. Otherwise it will be 0. This function aims to
    work similarly as its counterpart in numpy [1].

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._sign(device, input_image, output_image)

@plugin_function(categories=["combine", "binarize", "in assistant"])
def smaller(
    input_image0: Image,
    input_image1: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Determines if two images A and B smaller pixel wise. f(a, b) = 1 if a < b; 0
    otherwise.

    Parameters
    ----------
    input_image0: Image 
        First input image to process.
    input_image1: Image 
        Second input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_smaller
    """
    return clic._smaller(device, input_image0, input_image1, output_image)

@plugin_function(categories=["binarize", "in assistant"])
def smaller_constant(
    input_image: Image,
    output_image: Optional[Image] =None,
    scalar: float =0,
    device: Optional[Device] =None
) -> Image:
    """Determines if two images A and B smaller pixel wise. f(a, b) = 1 if a < b; 0
    otherwise.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    scalar: float (= 0)
        Scalar used in the comparison.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_smallerConstant
    """
    return clic._smaller_constant(device, input_image, output_image, float(scalar))

@plugin_function(categories=["combine", "binarize", "in assistant"])
def smaller_or_equal(
    input_image0: Image,
    input_image1: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Determines if two images A and B smaller or equal pixel wise. f(a, b) = 1 if a
    <= b; 0 otherwise.

    Parameters
    ----------
    input_image0: Image 
        First input image to process.
    input_image1: Image 
        Second input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_smallerOrEqual
    """
    return clic._smaller_or_equal(device, input_image0, input_image1, output_image)

@plugin_function(categories=["binarize", "in assistant"])
def smaller_or_equal_constant(
    input_image: Image,
    output_image: Optional[Image] =None,
    scalar: float =0,
    device: Optional[Device] =None
) -> Image:
    """Determines if two images A and B smaller or equal pixel wise. f(a, b) = 1 if a
    <= b; 0 otherwise.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    scalar: float (= 0)
        Scalar used in the comparison.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_smallerOrEqualConstant
    """
    return clic._smaller_or_equal_constant(device, input_image, output_image, float(scalar))

@plugin_function(categories=["filter", "edge detection", "in assistant", "bia-bob-suggestion"])
def sobel(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Convolve the image with the Sobel kernel. Author(s): Ruth WhelanJeans, Robert
    Haase

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_sobel
    """
    return clic._sobel(device, input_image, output_image)

@plugin_function(categories=["filter", "in assistant"])
def square_root(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Computes the square root of each pixel.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._square_root(device, input_image, output_image)

@plugin_function(categories=["projection", "in assistant", "bia-bob-suggestion"])
def std_z_projection(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Determines the standard deviation intensity projection of an image stack along
    Z.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_standardDeviationZProjection
    """
    return clic._std_z_projection(device, input_image, output_image)

@plugin_function(categories=["filter", "in assistant"])
def subtract_image_from_scalar(
    input_image: Image,
    output_image: Optional[Image] =None,
    scalar: float =0,
    device: Optional[Device] =None
) -> Image:
    """Subtracts one image X from a scalar s pixel wise. <pre>f(x, s) = s x</pre>

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    scalar: float (= 0)
        Scalar used in the subtraction.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_subtractImageFromScalar
    """
    return clic._subtract_image_from_scalar(device, input_image, output_image, float(scalar))

@plugin_function
def sum_reduction_x(
    input_image: Image,
    output_image: Optional[Image] =None,
    blocksize: int =256,
    device: Optional[Device] =None
) -> Image:
    """Takes an image and reduces it in width by factor blocksize. The new pixels
    contain the sum of the reduced pixels. For example, given the following image
    and block size 4: [0, 1, 1, 0, 1, 0, 1, 1] would lead to an image [2, 3]

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    blocksize: int (= 256)
        Blocksize value.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._sum_reduction_x(device, input_image, output_image, int(blocksize))

@plugin_function(categories=["projection"])
def sum_x_projection(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Determines the sum intensity projection of an image along Z.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_sumXProjection
    """
    return clic._sum_x_projection(device, input_image, output_image)

@plugin_function(categories=["projection"])
def sum_y_projection(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Determines the sum intensity projection of an image along Z.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_sumYProjection
    """
    return clic._sum_y_projection(device, input_image, output_image)

@plugin_function(categories=["projection", "in assistant", "bia-bob-suggestion"])
def sum_z_projection(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Determines the sum intensity projection of an image along Z.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_sumZProjection
    """
    return clic._sum_z_projection(device, input_image, output_image)

@plugin_function(categories=["transform"])
def transpose_xy(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Transpose X and Y axes of an image.

    Parameters
    ----------
    input_image: Image 
        The input image.
    output_image: Optional[Image] (= None)
        The output image where results are written into.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_transposeXY
    """
    return clic._transpose_xy(device, input_image, output_image)

@plugin_function(categories=["transform"])
def transpose_xz(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Transpose X and Z axes of an image.

    Parameters
    ----------
    input_image: Image 
        The input image.
    output_image: Optional[Image] (= None)
        The output image where results are written into.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_transposeXZ
    """
    return clic._transpose_xz(device, input_image, output_image)

@plugin_function(categories=["transform"])
def transpose_yz(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Transpose Y and Z axes of an image.

    Parameters
    ----------
    input_image: Image 
        The input image.
    output_image: Optional[Image] (= None)
        The output image where results are written into.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_transposeYZ
    """
    return clic._transpose_yz(device, input_image, output_image)

@plugin_function
def undefined_to_zero(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Copies all pixels instead those which are not a number (NaN) or infinity (inf),
    which are replaced by 0.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_undefinedToZero
    """
    return clic._undefined_to_zero(device, input_image, output_image)

@plugin_function(categories=["filter", "edge detection", "in assistant"])
def variance_box(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: int =1,
    radius_y: int =1,
    radius_z: int =1,
    device: Optional[Device] =None
) -> Image:
    """Computes the local variance of a pixels box neighborhood. The box size is
    specified by its halfwidth, halfheight and halfdepth (radius). If 2D images are
    given, radius_z will be ignored.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    radius_x: int (= 1)
        Radius size along x axis.
    radius_y: int (= 1)
        Radius size along y axis.
    radius_z: int (= 1)
        Radius size along z axis.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_varianceBox
    """
    return clic._variance_box(device, input_image, output_image, int(radius_x), int(radius_y), int(radius_z))

@plugin_function(categories=["filter", "edge detection", "in assistant"])
def variance_sphere(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: int =1,
    radius_y: int =1,
    radius_z: int =1,
    device: Optional[Device] =None
) -> Image:
    """Computes the local variance of a pixels sphere neighborhood. The sphere size is
    specified by its halfwidth, halfheight and halfdepth (radius). If 2D images are
    given, radius_z will be ignored.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    radius_x: int (= 1)
        Radius size along x axis.
    radius_y: int (= 1)
        Radius size along y axis.
    radius_z: int (= 1)
        Radius size along z axis.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_varianceSphere
    """
    return clic._variance_sphere(device, input_image, output_image, int(radius_x), int(radius_y), int(radius_z))

@plugin_function(categories=["filter", "edge detection", "in assistant"])
def variance(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: int =1,
    radius_y: int =1,
    radius_z: int =1,
    connectivity: str ="box",
    device: Optional[Device] =None
) -> Image:
    """Computes the local variance of a pixels neighborhood (box or sphere). The
    neighborhood size is specified by its halfwidth, halfheight and halfdepth
    (radius). If 2D images are given, radius_z will be ignored.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    radius_x: int (= 1)
        Radius size along x axis.
    radius_y: int (= 1)
        Radius size along y axis.
    radius_z: int (= 1)
        Radius size along z axis.
    connectivity: str (= "box")
        Filter neigborhood
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_varianceBox
    [2] https://clij.github.io/clij2-docs/reference_varianceSphere
    """
    return clic._variance(device, input_image, output_image, int(radius_x), int(radius_y), int(radius_z), str(connectivity))

@plugin_function
def write_values_to_positions(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Takes an image with three/four rows (2D: height = 3; 3D: height = 4): x, y [, z]
    and v and target image. The value v will be written at position x/y[/z] in the
    target image.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_writeValuesToPositions
    """
    return clic._write_values_to_positions(device, input_image, output_image)

@plugin_function(categories=["projection", "in assistant"])
def x_position_of_maximum_x_projection(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Determines an Xposition of the maximum intensity along X and writes it into the
    resulting image. If there are multiple xslices with the same value, the smallest
    X will be chosen.

    Parameters
    ----------
    input_image: Image 
        Input image stack
    output_image: Optional[Image] (= None)
        altitude map
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._x_position_of_maximum_x_projection(device, input_image, output_image)

@plugin_function(categories=["projection", "in assistant"])
def x_position_of_minimum_x_projection(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Determines an Xposition of the minimum intensity along X and writes it into the
    resulting image. If there are multiple xslices with the same value, the smallest
    X will be chosen.

    Parameters
    ----------
    input_image: Image 
        Input image stack
    output_image: Optional[Image] (= None)
        altitude map
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._x_position_of_minimum_x_projection(device, input_image, output_image)

@plugin_function(categories=["projection", "in assistant"])
def y_position_of_maximum_y_projection(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Determines an Yposition of the maximum intensity along Y and writes it into the
    resulting image. If there are multiple yslices with the same value, the smallest
    Y will be chosen.

    Parameters
    ----------
    input_image: Image 
        Input image stack
    output_image: Optional[Image] (= None)
        altitude map
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._y_position_of_maximum_y_projection(device, input_image, output_image)

@plugin_function(categories=["projection", "in assistant"])
def y_position_of_minimum_y_projection(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Determines an Yposition of the minimum intensity along Y and writes it into the
    resulting image. If there are multiple yslices with the same value, the smallest
    Y will be chosen.

    Parameters
    ----------
    input_image: Image 
        Input image stack
    output_image: Optional[Image] (= None)
        altitude map
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._y_position_of_minimum_y_projection(device, input_image, output_image)

@plugin_function(categories=["projection", "in assistant"])
def z_position_of_maximum_z_projection(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Determines a Zposition of the maximum intensity along Z and writes it into the
    resulting image. If there are multiple zslices with the same value, the smallest
    Z will be chosen.

    Parameters
    ----------
    input_image: Image 
        Input image stack
    output_image: Optional[Image] (= None)
        altitude map
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._z_position_of_maximum_z_projection(device, input_image, output_image)

@plugin_function(categories=["projection", "in assistant"])
def z_position_of_minimum_z_projection(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Determines a Zposition of the minimum intensity along Z and writes it into the
    resulting image. If there are multiple zslices with the same value, the smallest
    Z will be chosen.

    Parameters
    ----------
    input_image: Image 
        Input image stack
    output_image: Optional[Image] (= None)
        altitude map
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._z_position_of_minimum_z_projection(device, input_image, output_image)


import numpy as np

import pyclesperanto as cle


def test_affine_transform_translate():
    source = cle.push(
        np.asarray(
            [
                [
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                    [0, 0, 1, 1, 0],
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                ]
            ]
        )
    )

    reference = cle.push(
        np.asarray(
            [
                [
                    [0, 0, 0, 0, 0],
                    [0, 1, 1, 0, 0],
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                ]
            ]
        )
    )

    transform = [1, 0, -1, 0, 1, -1, 0, 0, 1]
    result = cle.affine_transform(source, transform_matrix=transform)

    a = cle.pull(result)
    b = cle.pull(reference)

    print(a)
    print(b)

    assert np.array_equal(a, b)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_setitem_single_value():
    image = cle.create([10, 20, 30])
    image[5, 10, 15] = 42

    result = cle.pull(image)
    assert result[5, 10, 15] == 42


def test_setitem_array():
    image = cle.create([10, 20, 30])
    data = np.ones((5, 10, 15))
    image[2:7, 5:15, 10:25] = data

    result = cle.pull(image)
    assert np.array_equal(result[2:7, 5:15, 10:25], data)


def test_setitem_invalid_value():
    image = cle.create([10, 20, 30])
    invalid_value = "invalid"

    try:
        image[5, 10, 15] = invalid_value
    except ValueError:
        pass
    else:
        assert False, "Expected TypeError to be raised"


def test_setitem_invalid_region():
    image = cle.create([10, 20, 30])
    data = np.ones((5, 10, 15))

    try:
        image[1, 5:6, 10:25] = data
    except IndexError:
        pass
    else:
        assert False, "Expected ValueError to be raised"


def test_setitem_out_of_bounds():
    image = cle.create([10, 20, 30])
    data = np.ones((5, 10, 15))

    try:
        image[15, 25, 35] = data
    except IndexError:
        pass
    else:
        assert False, "Expected IndexError to be raised"


def test_setitem_negative_indices():
    image = cle.create([10, 20, 30])
    data = np.ones((5, 10, 15))

    try:
        image[-5, -10, -15] = data
    except IndexError:
        pass
    else:
        assert False, "Expected IndexError to be raised"


def test_setitem_every_2_rows():
    image = cle.create([10, 20, 30])
    data = np.ones((5, 20, 30))
    image[::2, :, :] = data

    result = cle.pull(image)
    assert np.array_equal(result[::2, :, :], data)


def test_setitem_negative_step():
    image = cle.create([10, 20, 30])
    data = np.ones((5, 20, 30))
    image[2:7:-1, :, :] = data

    result = cle.pull(image)
    assert np.array_equal(result[2:7, :, :], data)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


input1 = np.asarray([[1, 2, 3]])
input2 = np.asarray([[4, 5, 7]])


def test_mean_squared_error():
    output = cle.mean_squared_error(input2, input1)
    assert abs(output - 11.333) < 0.001



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_subtract_image_from_scalar():
    test1 = cle.push(np.asarray([[0, 0], [1, 1], [2, 2]]))

    reference = cle.push(np.asarray([[5, 5], [4, 4], [3, 3]]))

    result = cle.create(test1)
    cle.subtract_image_from_scalar(test1, result, 5)

    a = cle.pull(result)
    b = cle.pull(reference)

    print(a)

    assert np.array_equal(a, b)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_flip():
    test = cle.push(
        np.asarray(
            [
                [0, 0, 0, 0, 0],
                [0, 1, 2, 0, 0],
                [0, 1, 2, 0, 0],
                [0, 1, 3, 0, 0],
                [0, 0, 0, 0, 0],
            ]
        )
    )

    reference = cle.push(
        np.asarray(
            [
                [0, 0, 0, 0, 0],
                [0, 0, 2, 1, 0],
                [0, 0, 2, 1, 0],
                [0, 0, 3, 1, 0],
                [0, 0, 0, 0, 0],
            ]
        )
    )

    result = cle.create(test)
    cle.flip(test, result, True, False, False)

    print(reference.get())
    print(result.get())

    a = cle.pull(result)
    b = cle.pull(reference)
    assert np.array_equal(a, b)



from typing import Optional, Union

import numpy as np

# from ._array import Array
from ._pyclesperanto import _Array as Array
from ._utils import (
    _assert_supported_dtype,
    _compute_range,
    _process_ellipsis_into_slice,
    _trim_index_to_shape,
)

cl_buffer_datatype_dict = {
    bool: "bool",
    np.uint8: "uchar",
    np.uint16: "ushort",
    np.uint32: "uint",
    np.uint64: "ulong",
    np.int8: "char",
    np.int16: "short",
    np.int32: "int",
    np.int64: "long",
    np.float32: "float",
    np.complex64: "cfloat_t",
    int: "int",
    float: "float",
    np.float64: "float",
}

_supported_numeric_types = tuple(cl_buffer_datatype_dict.keys())


def _astype(self, dtype: type):
    """Convert the Array to a different data type."""
    if dtype not in _supported_numeric_types:
        raise ValueError(
            "dtype "
            + str(dtype)
            + " not supported. Use one of "
            + str(_supported_numeric_types)
        )
    if dtype == self.dtype:
        return self

    from ._memory import create_like
    from ._tier1 import copy

    result = create_like(self, dtype=dtype)
    copy(input_image=self, output_image=result)
    return result


def _max(self, axis: Optional[int] = None, out=None):
    """Return the maximum value in the Array, or along an axis if specified."""
    from ._tier1 import maximum_x_projection, maximum_y_projection, maximum_z_projection
    from ._tier2 import maximum_of_all_pixels

    if axis == 0:
        result = maximum_z_projection(self)
    elif axis == 1:
        result = maximum_y_projection(self)
    elif axis == 2:
        result = maximum_x_projection(self)
    elif axis is None:
        result = maximum_of_all_pixels(self)
    else:
        raise ValueError("Axis " + str(axis) + " not supported")
    if out is not None:
        if isinstance(out, (Array, np.ndarray)):
            np.copyto(out, result.get().astype(out.dtype))
        else:
            out = result
    return result


def _min(self, axis: Optional[int] = None, out=None):
    """Return the minimum value in the Array, or along an axis if specified."""
    from ._tier1 import minimum_x_projection, minimum_y_projection, minimum_z_projection
    from ._tier2 import minimum_of_all_pixels

    if axis == 0:
        result = minimum_z_projection(self)
    elif axis == 1:
        result = minimum_y_projection(self)
    elif axis == 2:
        result = minimum_x_projection(self)
    elif axis is None:
        result = minimum_of_all_pixels(self)
    else:
        raise ValueError("Axis " + str(axis) + " not supported")
    if out is not None:
        if isinstance(out, (Array, np.ndarray)):
            np.copyto(out, result.get().astype(out.dtype))
    return result


def _sum(self, axis: Optional[int] = None, out=None):
    """Return the sum of the Array, or along an axis if specified."""
    from ._tier1 import sum_x_projection, sum_y_projection, sum_z_projection
    from ._tier2 import sum_of_all_pixels

    if axis == 0:
        result = sum_z_projection(self)
    elif axis == 1:
        result = sum_y_projection(self)
    elif axis == 2:
        result = sum_x_projection(self)
    elif axis is None:
        result = sum_of_all_pixels(self)
    else:
        raise ValueError("Axis " + str(axis) + " not supported")
    if out is not None:
        if isinstance(out, (Array, np.ndarray)):
            np.copyto(out, result.get().astype(out.dtype))
    return result


def __pos__(x1):
    """Unary plus, propagates the sign of the argument."""
    return x1.__mul__(1)


def __neg__(x1):
    """Unary minus, returns an array with the negative of the elements of the argument."""
    return x1.__mul__(-1)


def __add__(x1, x2):
    """Addition of two arrays."""
    if isinstance(x2, _supported_numeric_types):
        from ._tier1 import add_image_and_scalar

        return add_image_and_scalar(x1, scalar=x2)
    else:
        from ._tier1 import add_images_weighted

        return add_images_weighted(x1, x2, factor0=1, factor1=1)


def __iadd__(x1, x2):
    """Addition of two arrays."""
    from ._tier1 import copy

    temp = copy(x1)
    if isinstance(x2, _supported_numeric_types):
        from ._tier1 import add_image_and_scalar

        add_image_and_scalar(temp, output_image=x1, scalar=x2)
    else:
        from ._tier1 import add_images_weighted

        add_images_weighted(temp, x2, output_image=x1, factor0=1, factor1=1)
    return x1


def __sub__(x1, x2):
    """Subtraction of two arrays."""
    if isinstance(x2, _supported_numeric_types):
        from ._tier1 import add_image_and_scalar

        return add_image_and_scalar(x1, scalar=-x2)
    else:
        from ._tier1 import add_images_weighted

        return add_images_weighted(x1, x2, factor0=1, factor1=-1)


def __div__(x1, x2):
    """Division of two arrays."""
    if isinstance(x2, _supported_numeric_types):
        from ._tier1 import multiply_image_and_scalar

        return multiply_image_and_scalar(x1, scalar=1.0 / x2)
    else:
        from ._tier1 import divide_images

        return divide_images(x1, x2)


def __truediv__(x1, x2):
    """Division of two arrays."""
    return x1.__div__(x2)


def __idiv__(x1, x2):
    """Division of two arrays."""
    from ._tier1 import copy

    temp = copy(x1)
    if isinstance(x2, _supported_numeric_types):
        from ._tier1 import multiply_image_and_scalar

        return multiply_image_and_scalar(temp, x1, scalar=1.0 / x2)
    else:
        from ._tier1 import divide_images

        return divide_images(temp, x2, x1)


def __itruediv__(x1, x2):
    """Division of two arrays."""
    return x1.__idiv__(x2)


def __mul__(x1, x2):
    """Multiplication of two arrays."""
    if isinstance(x2, _supported_numeric_types):
        from ._tier1 import multiply_image_and_scalar

        return multiply_image_and_scalar(x1, scalar=x2)
    else:
        from ._tier1 import multiply_images

        return multiply_images(x1, x2)


def __imul__(x1, x2):
    """Multiplication of two arrays."""
    from ._tier1 import copy

    temp = copy(x1)
    if isinstance(x2, _supported_numeric_types):
        from ._tier1 import multiply_image_and_scalar

        return multiply_image_and_scalar(temp, x1, scalar=x2)
    else:
        from ._tier1 import multiply_images

        return multiply_images(temp, x2, x1)


def __gt__(x1, x2):
    """Greater than comparison of two arrays."""
    if isinstance(x2, _supported_numeric_types):
        from ._tier1 import greater_constant

        return greater_constant(x1, scalar=x2)
    else:
        from ._tier1 import greater

        return greater(x1, x2)


def __ge__(x1, x2):
    """Greater than or equal comparison of two arrays."""
    if isinstance(x2, _supported_numeric_types):
        from ._tier1 import greater_or_equal_constant

        return greater_or_equal_constant(x1, scalar=x2)
    else:
        from ._tier1 import greater_or_equal

        return greater_or_equal(x1, x2)


def __lt__(x1, x2):
    """Less than comparison of two arrays."""
    if isinstance(x2, _supported_numeric_types):
        from ._tier1 import smaller_constant

        return smaller_constant(x1, scalar=x2)
    else:
        from ._tier1 import smaller

        return smaller(x1, x2)


def __le__(x1, x2):
    """Less than or equal comparison of two arrays."""
    if isinstance(x2, _supported_numeric_types):
        from ._tier1 import smaller_or_equal_constant

        return smaller_or_equal_constant(x1, scalar=x2)
    else:
        from ._tier1 import smaller_or_equal

        return smaller_or_equal(x1, x2)


def __eq__(x1, x2):
    """Equal comparison of two arrays."""
    if isinstance(x2, _supported_numeric_types):
        from ._tier1 import equal_constant

        return equal_constant(x1, scalar=x2)
    else:
        from ._tier1 import equal

        return equal(x1, x2)


def __ne__(x1, x2):
    """Not equal comparison of two arrays."""
    if isinstance(x2, _supported_numeric_types):
        from ._tier1 import not_equal_constant

        return not_equal_constant(x1, scalar=x2)
    else:
        from ._tier1 import not_equal

        return not_equal(x1, x2)


def __pow__(x1, x2):
    """Power function of two arrays."""
    if isinstance(x2, _supported_numeric_types):
        from ._tier1 import power

        return power(x1, scalar=x2)
    else:
        from ._tier1 import power_images

        return power_images(x1, x2)


def __ipow__(x1, x2):
    """Power function of two arrays."""
    from ._tier1 import copy

    temp = copy(x1)
    if isinstance(x2, _supported_numeric_types):
        from ._tier1 import power

        return power(temp, x1, scalar=x2)
    else:
        from ._tier1 import power_images

        return power_images(temp, x2, x1)


def __iter__(self):
    """Iterate over the first dimension of the array."""

    class MyIterator:
        def __init__(self, image):
            self.image = image
            self._iter_index = 0

        def __next__(self):
            if not hasattr(self, "_iter_index"):
                self._iter_index = 0
            if self._iter_index < self.image.shape[0]:
                result = self.image[self._iter_index]
                self._iter_index = self._iter_index + 1
                return result
            else:
                raise StopIteration

    return MyIterator(self)


def __getitem__(self, index):
    """Get a pixel value or a region of interest from the Array."""
    if (
        isinstance(index, (tuple, list, np.ndarray))
        and index[0] is not None
        and isinstance(index[0], (tuple, list, np.ndarray))
    ):
        if len(index) == len(self.shape):
            if len(index[0]) > 0:
                # switch xy in 2D / xz in 3D, because clesperanto expects an X-Y-Z array;
                # see also https://github.com/clEsperanto/pyclesperanto_prototype/issues/49
                index = list(index)
                index[0], index[-1] = index[-1], index[0]

                # send positions to GPU
                from ._memory import push

                positions = push(np.asarray(index))

                # read values from positions
                from ._tier1 import read_values_from_positions

                return read_values_from_positions(self, positions)
            else:
                return []

    if not isinstance(index, tuple):
        index = (index,)

    if len(index) > self.ndim:
        raise IndexError(
            f"too many indices for array: array is {self.ndim}-dimensional, but {len(index)} were indexed"
        )

    dst_dim = sum(1 for i in index if not isinstance(i, (int, float)))

    index = _process_ellipsis_into_slice(index, self.shape)
    index = _trim_index_to_shape(index, self.shape)
    slice_list = [[0, x, 1] for x in self.shape]
    for x in range(len(index)):
        if isinstance(index[x], slice):
            slice_list[x] = [index[x].start, index[x].stop, index[x].step]
        elif np.issubdtype(type(index[x]), np.integer):
            slice_list[x] = [
                index[x],
                index[x] + 1 if index[x] >= 0 else index[x] - 1,
                None,
            ]

    # manage range for (x,y,z), with nothing that we deal with a z,y,x order
    _, range_x, range_y, range_z = _compute_range(slice_list, self.shape)
    origin = [range_z[0], range_y[0], range_x[0]]
    region = [
        range_z[1] - range_z[0],
        range_y[1] - range_y[0],
        range_x[1] - range_x[0],
    ]
    region = [abs(x) if x != 0 else 1 for x in region]
    steps = [range_z[2], range_y[2], range_x[2]]

    trimmed_region = [int(abs(x / s)) for x, s in zip(region, steps) if x > 1]
    dst_shape = [1] * dst_dim
    dst_shape[-len(trimmed_region) :] = trimmed_region

    # we are dealing with a single pixel operation
    if np.prod(region) == 1:
        # TODO: return a float or return a buffer of one?
        return self.get(origin, region)

    # a specific step was provided, we are dealing with a range operation
    from ._tier1 import range as gpu_range

    result = gpu_range(
        self,
        start_x=range_x[0],
        stop_x=range_x[1],
        step_x=range_x[2],
        start_y=range_y[0],
        stop_y=range_y[1],
        step_y=range_y[2],
        start_z=range_z[0],
        stop_z=range_z[1],
        step_z=range_z[2],
    )

    # if result is an Array, and one of the dimension is equal to 1
    if result.shape != tuple(dst_shape):
        from ._memory import create
        from ._tier1 import transpose_xy, transpose_yz

        tmp = create(
            dst_shape,
            dtype=self.dtype,
            mtype=self.mtype,
            device=self.device,
        )

        transpose = [x == 1 for x in region]
        if transpose[2]:  # x is empty
            result = transpose_xy(result)
            transpose_yz(result, tmp)
        elif transpose[1]:  # y is empty
            transpose_yz(result, tmp)
        else:
            result.copy(tmp)

        result = tmp

    return result


def __setitem__(self, index, value):
    """Set a pixel value or a region of interest in the Array."""
    if (
        isinstance(index, (tuple, np.ndarray))
        and index[0] is not None
        and isinstance(index[0], (tuple, list, np.ndarray))
    ):
        if len(index) == len(self.shape):
            if len(index[0]) > 0:
                # switch xy in 2D / xz in 3D, because clesperanto expects an X-Y-Z array;
                # see also https://github.com/clEsperanto/pyclesperanto_prototype/issues/49
                index = list(index)
                index[0], index[-1] = index[-1], index[0]

                # send positions to GPU
                from ._memory import push

                positions = push(np.asarray(index))

                num_positions = positions.shape[-1]
                if isinstance(value, (int, float)):
                    # make an array containing new values for every pixel
                    number = value

                    from ._memory import create

                    value_shape = [1] * len(self.shape)
                    value_shape[-1] = num_positions
                    value = create(value_shape)
                    value.fill(number)

                # overwrite pixels
                from ._tier1 import write_values_to_positions
                from ._tier2 import concatenate_along_y

                values_and_positions = concatenate_along_y(positions, value)
                write_values_to_positions(values_and_positions, self)
            return

    if not isinstance(value, (Array, np.ndarray)):
        value = np.array(value)

    if not isinstance(index, tuple):
        index = (index,)

    dst_dim = sum(1 for i in index if not isinstance(i, (int, float)))

    index = _process_ellipsis_into_slice(index, self.shape)
    index = _trim_index_to_shape(index, self.shape)
    slice_list = [[0, x, 1] for x in self.shape]
    for x in range(len(index)):
        if isinstance(index[x], slice):
            slice_list[x] = [index[x].start, index[x].stop, index[x].step]
        elif np.issubdtype(type(index[x]), np.integer):
            slice_list[x] = [
                index[x],
                index[x] + 1 if index[x] >= 0 else index[x] - 1,
                None,
            ]

    # manage range for (x,y,z), with nothing that we deal with a z,y,x order
    _, range_x, range_y, range_z = _compute_range(slice_list, self.shape)
    origin = [range_z[0], range_y[0], range_x[0]]
    region = [
        range_z[1] - range_z[0],
        range_y[1] - range_y[0],
        range_x[1] - range_x[0],
    ]
    region = [abs(x) if x != 0 else 1 for x in region]
    steps = [range_z[2], range_y[2], range_x[2]]

    trimmed_region = [int(abs(x / s)) for x, s in zip(region, steps) if x > 1]
    dst_shape = [1] * dst_dim
    dst_shape[-len(trimmed_region) :] = trimmed_region

    if value.size == 1:
        if np.prod(region) > 1:
            value = np.repeat(value, np.prod(region))
            value = value.reshape(region)

        self.set(value, origin, region)
        return

    if any([x != 1 for x in steps]):
        # TODO: not sure it work properly
        from ._tier1 import range as gpu_range

        gpu_range(
            value,
            self,
            start_x=range_x[0],
            stop_x=range_x[1],
            step_x=range_x[2],
            start_y=range_y[0],
            stop_y=range_y[1],
            step_y=range_y[2],
            start_z=range_z[0],
            stop_z=range_z[1],
            step_z=range_z[2],
        )
        return

    if isinstance(value, Array):
        if self.dtype == value.dtype:
            self.copy(value, origin, (0, 0, 0), region)
        else:
            # otherwise we copy with cast using paste
            from ._tier1 import paste

            paste(
                value,
                self,
                index_x=origin[-1] if len(origin) > 0 else 0,
                index_y=origin[-2] if len(origin) > 1 else 0,
                index_z=origin[-3] if len(origin) > 2 else 0,
            )
    else:
        self.set(value, origin, region)


# adapted from https://github.com/napari/napari/blob/d6bc683b019c4a3a3c6e936526e29bbd59cca2f4/napari/utils/notebook_display.py#L54-L73
def __plt_to_png__(self):
    """PNG representation of the image object for IPython.
    Returns
    -------
    In memory binary stream containing a PNG matplotlib image.
    """
    from io import BytesIO

    import matplotlib.pyplot as plt

    with BytesIO() as file_obj:
        plt.savefig(file_obj, format="png")
        plt.close()  # supress plot output
        file_obj.seek(0)
        png = file_obj.read()
    return png


def __png_to_html__(self, png):
    import base64

    url = "data:image/png;base64," + base64.b64encode(png).decode("utf-8")
    return f'<img src="{url}"></img>'


def __repr_html__(self):
    """HTML representation of the image object for IPython.
    Returns
    -------
    HTML text with the image and some properties.
    """
    import matplotlib.pyplot as plt
    import numpy as np

    from ._functionalities import imshow

    size_in_pixels = np.prod(self.size)
    size_in_bytes = size_in_pixels * self.dtype.itemsize

    labels = self.dtype == np.uint32

    # In case the image is 2D, 3D and larger than 100 pixels, turn on fancy view
    if len(self.shape) in (2, 3) and size_in_pixels >= 100:
        imshow(self, labels=labels, continue_drawing=True, colorbar=not labels)
        image = self._png_to_html(self._plt_to_png())
    else:
        return "<pre>" + repr(self) + "</pre>"

    units = ["B", "kB", "MB", "GB", "TB", "PB"]
    unit_index = 0
    while size_in_bytes > 1024 and unit_index < len(units) - 1:
        size_in_bytes /= 1024
        unit_index += 1
    size = "{:.1f}".format(size_in_bytes) + " " + units[unit_index]

    histogram = ""
    if size_in_bytes < 100 * 1024 * 1024:
        if not labels:
            from ._tier3 import histogram

            num_bins = 32
            h = np.asarray(
                histogram(self, nbins=num_bins, min=self.min(), max=self.max())
            )
            plt.figure(figsize=(1.8, 1.2))
            plt.bar(range(0, len(h)), h)
            # hide axis text
            # https://stackoverflow.com/questions/2176424/hiding-axis-text-in-matplotlib-plots
            # https://pythonguides.com/matplotlib-remove-tick-labels
            frame1 = plt.gca()
            frame1.axes.xaxis.set_ticklabels([])
            frame1.axes.yaxis.set_ticklabels([])
            plt.tick_params(left=False, bottom=False)
            histogram = self._png_to_html(self._plt_to_png())
        min_max = (
            "<tr><td>min</td><td>"
            + str(self.min())
            + "</td></tr>"
            + "<tr><td>max</td><td>"
            + str(self.max())
            + "</td></tr>"
        )
    else:
        min_max = ""
    all = [
        "<table>",
        "<tr>",
        "<td>",
        image,
        "</td>",
        '<td style="text-align: center; vertical-align: top;">',
        '<b><a href="https://github.com/clEsperanto/pyclesperanto_prototype" target="_blank">cle._</a> image</b><br/>',
        "<table>",
        "<tr><td>shape</td><td>"
        + str(self.shape).replace(" ", "&nbsp;")
        + "</td></tr>",
        "<tr><td>dtype</td><td>" + str(self.dtype) + "</td></tr>",
        "<tr><td>size</td><td>" + size + "</td></tr>",
        min_max,
        "</table>",
        histogram,
        "</td>",
        "</tr>",
        "</table>",
    ]
    return "\n".join(all)



#
# This code is auto-generated from CLIc 'cle::tier8.hpp' file, do not edit manually.
#

import importlib
import warnings
from typing import Optional

import numpy as np

from ._array import Image
from ._core import Device
from ._decorators import plugin_function

clic = importlib.import_module('._pyclesperanto', package='pyclesperanto')

@plugin_function(categories=["label processing", "in assistant", "bia-bob-suggestion"])
def smooth_labels(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius: int =0,
    device: Optional[Device] =None
) -> Image:
    """Apply a morphological opening operation to a label image and afterwards   fills
    gaps between the labels using voronoi-labeling. Finally, the result   label
    image is masked so that all background pixels remain background pixels.   Note:
    It is recommended to process isotropic label images.

    Parameters
    ----------
    input_image: Image 
        Input label image
    output_image: Optional[Image] (= None)
        Output label image
    radius: int (= 0)
        Smoothing
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._smooth_labels(device, input_image, output_image, int(radius))

@plugin_function(categories=["label processing", "in assistant", "bia-bob-suggestion"])
def smooth_connected_labels(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius: int =0,
    device: Optional[Device] =None
) -> Image:
    """Apply a morphological erosion and dilation of the label image with respect to
    the connectivity of the labels.   Note: It is recommended to process isotropic
    label images.

    Parameters
    ----------
    input_image: Image 
        Input label image
    output_image: Optional[Image] (= None)
        Output label image
    radius: int (= 0)
        Smoothing
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._smooth_connected_labels(device, input_image, output_image, int(radius))